@@ -1,21 +1,22 @@
 /* [argparse.c wk 17.06.97] Argument Parser for option handling
- *	Copyright (C) 1998, 1999, 2000, 2001 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 1999, 2000, 2001, 2006 Free Software Foundation, Inc.
  *
- *  This file is part of GnuPG.
+ * This file is part of JNLIB.
  *
- *  GnuPG is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ * JNLIB is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
  *
- *  GnuPG is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ * JNLIB is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.
  */
 
 #include <config.h>
@@ -819,8 +820,20 @@ show_help( ARGPARSE_OPTS *opts, unsigned flags )
 	    puts("\n(A single dash may be used instead of the double ones)");
     }
     if( (s=strusage(19)) ) {  /* bug reports to ... */
+        /* To allow changing the bug reporting address without
+         * touching an i18n string, we replace a placeholder @EMAIL@
+         * while printing by a string given with strusage(49). */
+        const char *email, *point;
 	putchar('\n');
-	fputs(s, stdout);
+        email = strusage (49);
+        if (email && *email && (point = strstr (s, "@EMAIL@")) ) {
+            for (; s < point; s++)
+                putchar (*s);
+            fputs (email, stdout);
+            s = point + 7;
+        }
+        fputs(s, stdout);
+        
     }
     fflush(stdout);
     exit(0);
@@ -852,7 +865,7 @@ show_version()
     /* additional program info */
     for(i=30; i < 40; i++ )
 	if( (s=strusage(i)) )
-	    fputs( (const byte*)s, stdout);
+	    fputs (s, stdout);
     fflush(stdout);
 }
 
@@ -887,11 +900,13 @@ usage( int level )
  *    16: Long copying conditions (with LFs)
  *    17: Optional printable OS name
  *    18: Optional thanks list	 (with LFs)
- *    19: Bug report info
+ *    19: Bug report info (see also 190).
  *20..29: Additional lib version strings.
  *30..39: Additional program info (with LFs)
  *    40: short usage note (with LF)
  *    41: long usage note (with LF)
+ *    49: bug reporting address: If not empty, insert this string 
+ *        into the string returned by 19 at @EMAIL@
  */
 const char *
 strusage( int level )
@@ -904,7 +919,7 @@ strusage( int level )
     switch( level ) {
       case 11: p = "foo"; break;
       case 13: p = "0.0"; break;
-      case 14: p = "Copyright (C) 2005 Free Software Foundation, Inc."; break;
+      case 14: p = "Copyright (C) 2006 Free Software Foundation, Inc."; break;
       case 15: p =
 "This program comes with ABSOLUTELY NO WARRANTY.\n"
 "This is free software, and you are welcome to redistribute it\n"
@@ -920,7 +935,8 @@ strusage( int level )
 "GNU General Public License for more details.\n\n"
 "You should have received a copy of the GNU General Public License\n"
 "along with this program; if not, write to the Free Software\n"
-"Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n";
+"Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,\n"
+"USA.\n";
 	break;
       case 40: /* short and long usage */
       case 41: p = ""; break;
