@@ -117,7 +117,7 @@ compare_serialno (ksba_sexp_t serial1, ksba_sexp_t serial2 )
   unsigned long n_a, n_b;
 
   if (!a || !b)
-    return (!a && !b);
+    return !(!a && !b);
 
   n_a = strtoul (a+1, (char**)&endp, 10);
   a = endp;
@@ -126,10 +126,10 @@ compare_serialno (ksba_sexp_t serial1, ksba_sexp_t serial2 )
   if (*a != ':' || *b != ':')
     {
       log_debug ("invalid S-exp in compare_serialno\n");
-      return 0;
+      return 1;
     }
   if (n_a != n_b)    
-    return 0;
+    return 1;
   a++;
   b++;
 
@@ -199,13 +199,19 @@ clean_cache_slot (cert_item_t ci)
 
 
 /* Put the certificate CERT into the cache.  It is assumed that the
-   cache is locked while this function is called. */
+   cache is locked while this function is called. If FPR_BUFFER is not
+   NULL the fingerprint of the certificate will be stored there.
+   FPR_BUFFER neds to point to a buffer of at least 20 bytes. The
+   fingerprint will be stored on success or when the function returns
+   gpg_err_code(GPG_ERR_DUP_VALUE). */
 static gpg_error_t
-put_cert (ksba_cert_t cert, int is_loaded, int is_trusted)
+put_cert (ksba_cert_t cert, int is_loaded, int is_trusted, void *fpr_buffer)
 {
-  unsigned char fpr[20];
+  unsigned char help_fpr_buffer[20], *fpr;
   cert_item_t ci;
 
+  fpr = fpr_buffer? fpr_buffer : &help_fpr_buffer;
+
   /* If we already reached the caching limit, drop a couple of certs
      from the cache.  Our dropping strategy is simple: We keep a
      static index counter and use this to start looking for
@@ -360,23 +366,28 @@ load_certs_from_dir (const char *dirname, int are_trusted)
           continue;
         }
 
-      err = put_cert (cert, 1, are_trusted);
-      ksba_cert_release (cert);
+      err = put_cert (cert, 1, are_trusted, NULL);
       if (gpg_err_code (err) == GPG_ERR_DUP_VALUE)
         log_info (_("certificate `%s' already cached\n"), fname);
       else if (!err)
         {
-          log_info (_("certificate `%s' loaded\n"), fname);
-          if (are_trusted || opt.verbose)
+          if (are_trusted)
+            log_info (_("trusted certificate `%s' loaded\n"), fname);
+          else
+            log_info (_("certificate `%s' loaded\n"), fname);
+          if (opt.verbose)
             {
               p = get_fingerprint_hexstring_colon (cert);
-              log_info (_("SHA1 fingerprint = %s\n"), p);
+              log_info (_("  SHA1 fingerprint = %s\n"), p);
               xfree (p);
+
+              cert_log_name (_("  name ="), cert);
             }
         }
       else
         log_error (_("error loading certificate `%s': %s\n"),
                      fname, gpg_strerror (err));
+      ksba_cert_release (cert);
     }
 
   xfree (fname);
@@ -463,19 +474,37 @@ cache_cert (ksba_cert_t cert)
   gpg_error_t err;
 
   acquire_cache_write_lock ();
-  err = put_cert (cert, 0, 0);
+  err = put_cert (cert, 0, 0, NULL);
   release_cache_lock ();
   if (gpg_err_code (err) == GPG_ERR_DUP_VALUE)
     log_info (_("certificate already cached\n"));
   else if (!err)
     log_info (_("certificate cached\n"));
   else
-    log_error (_("error caching certificate: %s\n"),
-               gpg_strerror (err));
+    log_error (_("error caching certificate: %s\n"), gpg_strerror (err));
   return err;
 }
 
 
+/* Put CERT into the certificate cache and store the fingerprint of
+   the certificate into FPR_BUFFER.  If the certificate is already in
+   the cache do not print a warning; just store the
+   fingerprint. FPR_BUFFER needs to be at least 20 bytes. */
+gpg_error_t
+cache_cert_silent (ksba_cert_t cert, void *fpr_buffer)
+{
+  gpg_error_t err;
+
+  acquire_cache_write_lock ();
+  err = put_cert (cert, 0, 0, fpr_buffer);
+  release_cache_lock ();
+  if (gpg_err_code (err) == GPG_ERR_DUP_VALUE)
+    err = 0;
+  if (err)
+    log_error (_("error caching certificate: %s\n"), gpg_strerror (err));
+  return err;
+}
+
 
 
 /* Return a certificate object for the given fingerprint.  FPR is
@@ -626,48 +655,52 @@ get_cert_bysubject (const char *subject_dn, unsigned int seq)
 
 
 
-
 /* Return the certificate matching ISSUER_DN and SERIALNO; if it is
    not already in the cache, try to find it from other resources.  */
 ksba_cert_t
 find_cert_bysn (ctrl_t ctrl, const char *issuer_dn, ksba_sexp_t serialno)
 {
   gpg_error_t err;
-  ksba_cert_t cert = NULL;
+  ksba_cert_t cert;
   cert_fetch_context_t context = NULL;
+  char *hexsn, *buf;
 
+  /* First check whether it has already been cached.  */
   cert = get_cert_bysn (issuer_dn, serialno);
+  if (cert)
+    return cert;
 
-  if (!cert)
+  /* Ask back to the service requester to return the certificate.
+     This is because we can assume that he already used the
+     certificate while checking for the CRL. */
+  hexsn = serial_hex (serialno);
+  if (!hexsn)
     {
-      char *hexsn, *buf;
-      /* Ask back to the service requester to return the certificate.
-         This is because we can assume that he already used the
-         certificate while checking for the CRL. */
-      hexsn = serial_hex (serialno);
-      if (!hexsn)
-        {
-          log_debug ("serial_hex() failed\n");
-          return NULL;
-        }
-      buf = xtrymalloc (1 + strlen (hexsn) + 1 + strlen (issuer_dn) + 1);
-      if (!buf)
-        {
-          log_error ("can't allocate enough memory: %s\n", strerror (errno));
-          xfree (hexsn);
-          return NULL;
-        }
-      strcpy (stpcpy (stpcpy (stpcpy (buf, "#"), hexsn),"/"), issuer_dn);
+      log_error ("serial_hex() failed\n");
+      return NULL;
+    }
+  buf = xtrymalloc (1 + strlen (hexsn) + 1 + strlen (issuer_dn) + 1);
+  if (!buf)
+    {
+      log_error ("can't allocate enough memory: %s\n", strerror (errno));
       xfree (hexsn);
-      cert = get_cert_local (ctrl, buf);
-      xfree (buf);
-      if (cert)
-        {
-          cache_cert (cert);
-          return cert; /* Done. */
-        }
+      return NULL;
+    }
+  strcpy (stpcpy (stpcpy (stpcpy (buf, "#"), hexsn),"/"), issuer_dn);
+  xfree (hexsn);
+  cert = get_cert_local (ctrl, buf);
+  xfree (buf);
+  if (cert)
+    {
+      cache_cert (cert);
+      return cert; /* Done. */
     }
 
+  if (DBG_LOOKUP)
+    log_debug ("find_cert_bysn: certificate not returned by caller"
+               " - doing lookup\n");
+
+  /* Retrieve the certificate from external resources. */
   while (!cert)
     {
       ksba_sexp_t sn;
@@ -716,6 +749,7 @@ find_cert_bysn (ctrl_t ctrl, const char *issuer_dn, ksba_sexp_t serialno)
       if (!compare_serialno (serialno, sn))
         {
           ksba_free (sn);
+          ksba_free (issdn);
           cache_cert (cert);
           if (DBG_LOOKUP)
             log_debug ("   found\n");
@@ -723,7 +757,7 @@ find_cert_bysn (ctrl_t ctrl, const char *issuer_dn, ksba_sexp_t serialno)
         }
 
       ksba_free (sn);
-
+      ksba_free (issdn);
       ksba_cert_release (cert);
       cert = NULL;
     }
@@ -732,6 +766,7 @@ find_cert_bysn (ctrl_t ctrl, const char *issuer_dn, ksba_sexp_t serialno)
   return cert;
 }
 
+
 /* Return the certificate matching SUBJECT_DN and (if not NULL)
    KEYID. If it is not already in the cache, try to find it from other
    resources.  Note, that the external search does not work for user
@@ -746,6 +781,38 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
   cert_fetch_context_t context = NULL;
   ksba_sexp_t subj;
 
+  /* If we have certificates from an OCSP request we first try to use
+     them.  This is because these certificates will really be the
+     required ones and thus even in the case that they can't be
+     uniquely located by the following code we can use them.  This is
+     for example required by Telesec certificates where a keyId is
+     used but the issuer certificate comes without a subject keyId! */
+  if (ctrl->ocsp_certs)
+    {
+      cert_item_t ci;
+      cert_ref_t cr;
+      int i;
+
+      /* For efficiency reasons we won't use get_cert_bysubject here. */
+      acquire_cache_read_lock ();
+      for (i=0; i < 256; i++)
+        for (ci=cert_cache[i]; ci; ci = ci->next)
+          if (ci->cert && ci->subject_dn
+              && !strcmp (ci->subject_dn, subject_dn))
+            for (cr=ctrl->ocsp_certs; cr; cr = cr->next)
+              if (!memcmp (ci->fpr, cr->fpr, 20))
+                {
+                  ksba_cert_ref (ci->cert);
+                  release_cache_lock ();
+                  return ci->cert; /* We use this certificate. */
+                }
+      release_cache_lock ();
+      if (DBG_LOOKUP)
+        log_debug ("find_cert_bysubject: certificate not in ocsp_certs\n");
+    }
+
+
+  /* First we check whether the certificate is cached.  */
   for (seq=0; (cert = get_cert_bysubject (subject_dn, seq)); seq++)
     {
       if (!keyid)
@@ -759,21 +826,45 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
       xfree (subj);
       ksba_cert_release (cert);
     }
+  if (cert)
+    return cert; /* Done.  */
 
-  if (!cert)
+  if (DBG_LOOKUP)
+    log_debug ("find_cert_bysubject: certificate not in cache\n");
+
+  /* Ask back to the service requester to return the certificate.
+     This is because we can assume that he already used the
+     certificate while checking for the CRL. */
+  if (keyid)
+    cert = get_cert_local_ski (ctrl, subject_dn, keyid);
+  else
     {
-      /* Ask back to the service requester to return the certificate.
-         This is because we can assume that he already used the
-         certificate while checking for the CRL. */
-      cert = get_cert_local_ski (ctrl, subject_dn, keyid);
-      if (cert)
+      /* In contrast to get_cert_local_ski, get_cert_local uses any
+         passed pattern, so we need to make sure that an exact subject
+         search is done. */
+      char *buf;
+
+      buf = xtrymalloc (1 + strlen (subject_dn) + 1);
+      if (!buf)
         {
-          cache_cert (cert);
-          return cert; /* Done. */
+          log_error ("can't allocate enough memory: %s\n", strerror (errno));
+          return NULL;
         }
+      strcpy (stpcpy (buf, "/"), subject_dn);
+      cert = get_cert_local (ctrl, buf);
+      xfree (buf);
+    }
+  if (cert)
+    {
+      cache_cert (cert);
+      return cert; /* Done. */
     }
 
+  if (DBG_LOOKUP)
+    log_debug ("find_cert_bysubject: certificate not returned by caller"
+               " - doing lookup\n");
 
+  /* Locate the certificate using external resources. */
   while (!cert)
     {
       char *subjdn;
@@ -816,6 +907,7 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
         }
       ksba_free (subjdn);
 
+      /* If no key ID has been provided, we return the first match.  */
       if (!keyid)
         {
           cache_cert (cert);
@@ -824,6 +916,7 @@ find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
           break; /* Ready.  */
         }
 
+      /* With the key ID given we need to compare it.  */
       if (!ksba_cert_get_subj_key_id (cert, NULL, &subj))
         {
           if (!cmp_simple_canon_sexp (keyid, subj))
@@ -953,6 +1046,8 @@ find_issuing_cert (ctrl_t ctrl, ksba_cert_t cert, ksba_cert_t *r_cert)
   if (err || !issuer_cert)
     {
       issuer_cert = get_cert_bysubject (issuer_dn, 0);
+      if (issuer_cert)
+        err = 0;
     }
 
  leave:  
