@@ -137,11 +137,10 @@ my_strusage (int level)
     case 11: p = "dirmngr-client (dirmngr)";
       break;
     case 13: p = VERSION; break;
-    case 14: p = "Copyright (C) 2004 g10 Code GmbH"; break;
+    case 14: p = "Copyright (C) " COPYRIGHT_YEAR_NAME; break;
     case 17: p = PRINTABLE_OS_NAME; break;
-    case 19: p =
-               _("Please report bugs to <gpa-dev@gnupg.org>.\n");
-      break;
+    case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
+    case 49: p = PACKAGE_BUGREPORT; break;
     case 1:
     case 40: p =
                  _("Usage: dirmngr-client [options] "
@@ -197,6 +196,7 @@ main (int argc, char **argv )
   /* Init Assuan.  */
   assuan_set_assuan_log_stream (log_get_stream ());
   assuan_set_assuan_log_prefix (log_get_prefix (NULL));
+  assuan_set_assuan_err_source (GPG_ERR_SOURCE_DEFAULT);
 
   /* Setup I18N. */
   i18n_init();
@@ -399,7 +399,7 @@ main (int argc, char **argv )
 
 
 /* Print status line from the assuan protocol.  */
-static assuan_error_t
+static int
 status_cb (void *opaque, const char *line)
 {
   if (opt.verbose > 2)
@@ -408,7 +408,7 @@ status_cb (void *opaque, const char *line)
 }
 
 /* Print data as retrieved by the lookup function.  */
-static assuan_error_t
+static int
 data_cb (void *opaque, const void *buffer, size_t length)
 {
   gpg_error_t err;
@@ -475,7 +475,7 @@ start_dirmngr (int only_daemon)
       no_close_list[i] = -1;
 
       /* Connect to the agent and perform initial handshaking.  */
-      rc = assuan_pipe_connect (&ctx, opt.dirmngr_program, (char**)argv,
+      rc = assuan_pipe_connect (&ctx, opt.dirmngr_program, argv,
                                 no_close_list);
     }
   else /* Connect to a daemon.  */
@@ -517,7 +517,7 @@ start_dirmngr (int only_daemon)
 
       rc = assuan_socket_connect (&ctx, infostr, pid);
       xfree (infostr);
-      if (rc == ASSUAN_Connect_Failed && !only_daemon)
+      if (gpg_err_code(rc) == GPG_ERR_ASS_CONNECT_FAILED && !only_daemon)
         {
           log_error (_("can't connect to the dirmngr - trying fall back\n"));
           opt.force_pipe_server = 1;
@@ -528,7 +528,7 @@ start_dirmngr (int only_daemon)
   if (rc)
     {
       log_error (_("can't connect to the dirmngr: %s\n"),
-                 assuan_strerror (rc));
+                 gpg_strerror (rc));
       return NULL;
     }
 
@@ -744,36 +744,47 @@ read_certificate (const char *fname, unsigned char **rbuf, size_t *rbuflen)
 
 
 /* Callback for the inquire fiunction to send back the certificate.  */
-static assuan_error_t
+static int
 inq_cert (void *opaque, const char *line)
 {
   struct inq_cert_parm_s *parm = opaque;
-  assuan_error_t ae;
+  gpg_error_t err;
 
   if (!strncmp (line, "TARGETCERT", 10) && (line[10] == ' ' || !line[10]))
     {
-      ae = assuan_send_data (parm->ctx, parm->cert, parm->certlen);
+      err = assuan_send_data (parm->ctx, parm->cert, parm->certlen);
     }
   else if (!strncmp (line, "SENDCERT", 8) && (line[8] == ' ' || !line[8]))
     {
       /* We don't support this but dirmngr might ask for it.  So
          simply ignore it by sending back and empty value. */
-      ae = assuan_send_data (parm->ctx, NULL, 0);
+      err = assuan_send_data (parm->ctx, NULL, 0);
+    }
+  else if (!strncmp (line, "SENDCERT_SKI", 12)
+           && (line[12]==' ' || !line[12]))
+    {
+      /* We don't support this but dirmngr might ask for it.  So
+         simply ignore it by sending back an empty value. */
+      err = assuan_send_data (parm->ctx, NULL, 0);
     }
   else if (!strncmp (line, "SENDISSUERCERT", 14)
            && (line[14] == ' ' || !line[14]))
     {
       /* We don't support this but dirmngr might ask for it.  So
-         simply ignore it by sending back and empty value. */
-      ae = assuan_send_data (parm->ctx, NULL, 0);
+         simply ignore it by sending back an empty value. */
+      err = assuan_send_data (parm->ctx, NULL, 0);
     }
   else
     {
       log_info (_("unsupported inquiry `%s'\n"), line);
-      ae = ASSUAN_Inquire_Unknown;
+      err = gpg_error (GPG_ERR_ASS_UNKNOWN_INQUIRE);
+      /* Note that this error will let assuan_transact terminate
+         immediately instead of return the error to the caller.  It is
+         not clear whether this is the desired behaviour - it may
+         change in future. */
     }
 
-  return ae;
+  return err;
 }
 
 
@@ -782,7 +793,7 @@ inq_cert (void *opaque, const char *line)
 static gpg_error_t
 do_check (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
 {
-  assuan_error_t ae;
+  gpg_error_t err;
   struct inq_cert_parm_s parm;
 
   memset (&parm, 0, sizeof parm);
@@ -790,12 +801,12 @@ do_check (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
   parm.cert = cert;
   parm.certlen = certlen;
 
-  ae = assuan_transact (ctx, opt.use_ocsp? "CHECKOCSP":"CHECKCRL", NULL, NULL,
+  err = assuan_transact (ctx, opt.use_ocsp? "CHECKOCSP":"CHECKCRL", NULL, NULL,
                         inq_cert, &parm,
                         status_cb, NULL);
   if (opt.verbose > 1)
-    log_info ("response of dirmngr: %s\n", ae? assuan_strerror (ae): "okay");
-  return map_assuan_err (ae);
+    log_info ("response of dirmngr: %s\n", err? gpg_strerror (err): "okay");
+  return err;
 }
 
 /* Check the certificate CERT,CERTLEN for validity using a CRL or OCSP.
@@ -803,7 +814,7 @@ do_check (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
 static gpg_error_t
 do_cache (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
 {
-  assuan_error_t ae;
+  gpg_error_t err;
   struct inq_cert_parm_s parm;
 
   memset (&parm, 0, sizeof parm);
@@ -811,12 +822,12 @@ do_cache (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
   parm.cert = cert;
   parm.certlen = certlen;
 
-  ae = assuan_transact (ctx, "CACHECERT", NULL, NULL,
+  err = assuan_transact (ctx, "CACHECERT", NULL, NULL,
                         inq_cert, &parm,
                         status_cb, NULL);
   if (opt.verbose > 1)
-    log_info ("response of dirmngr: %s\n", ae? assuan_strerror (ae): "okay");
-  return map_assuan_err (ae);
+    log_info ("response of dirmngr: %s\n", err? gpg_strerror (err): "okay");
+  return err;
 }
 
 /* Check the certificate CERT,CERTLEN for validity using dirmngrs
@@ -824,7 +835,7 @@ do_cache (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
 static gpg_error_t
 do_validate (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
 {
-  assuan_error_t ae;
+  gpg_error_t err;
   struct inq_cert_parm_s parm;
 
   memset (&parm, 0, sizeof parm);
@@ -832,19 +843,19 @@ do_validate (assuan_context_t ctx, const unsigned char *cert, size_t certlen)
   parm.cert = cert;
   parm.certlen = certlen;
 
-  ae = assuan_transact (ctx, "VALIDATE", NULL, NULL,
+  err = assuan_transact (ctx, "VALIDATE", NULL, NULL,
                         inq_cert, &parm,
                         status_cb, NULL);
   if (opt.verbose > 1)
-    log_info ("response of dirmngr: %s\n", ae? assuan_strerror (ae): "okay");
-  return map_assuan_err (ae);
+    log_info ("response of dirmngr: %s\n", err? gpg_strerror (err): "okay");
+  return err;
 }
 
 /* Load a CRL into the dirmngr.  */
 static gpg_error_t
 do_loadcrl (assuan_context_t ctx, const char *filename)
 {
-  assuan_error_t ae;
+  gpg_error_t err;
   const char *s;
   char *fname, *line, *p;
 
@@ -881,14 +892,14 @@ do_loadcrl (assuan_context_t ctx, const char *filename)
     }
   *p = 0;
 
-  ae = assuan_transact (ctx, line, NULL, NULL,
+  err = assuan_transact (ctx, line, NULL, NULL,
                         NULL, NULL,
                         status_cb, NULL);
   if (opt.verbose > 1)
-    log_info ("response of dirmngr: %s\n", ae? assuan_strerror (ae): "okay");
+    log_info ("response of dirmngr: %s\n", err? gpg_strerror (err): "okay");
   xfree (line);
   xfree (fname);
-  return map_assuan_err (ae);
+  return err;
 }
 
 
@@ -898,7 +909,6 @@ static gpg_error_t
 do_lookup (assuan_context_t ctx, const char *pattern)
 {
   gpg_error_t err;
-  assuan_error_t ae;
   const unsigned char *s;
   char *line, *p;
   struct b64state state;
@@ -928,17 +938,17 @@ do_lookup (assuan_context_t ctx, const char *pattern)
   *p = 0;
 
 
-  ae = assuan_transact (ctx, line,
-                        data_cb, &state,
-                        NULL, NULL,
-                        status_cb, NULL);
+  err = assuan_transact (ctx, line,
+                         data_cb, &state,
+                         NULL, NULL,
+                         status_cb, NULL);
   if (opt.verbose > 1)
-    log_info ("response of dirmngr: %s\n", ae? assuan_strerror (ae): "okay");
+    log_info ("response of dirmngr: %s\n", err? gpg_strerror (err): "okay");
 
   err = b64enc_finish (&state);
 
   xfree (line);
-  return ae? map_assuan_err (ae) : err;
+  return err;
 }
 
 /* The body of an endless loop: Read a line from stdin, retrieve the
