@@ -1,6 +1,6 @@
 /* dirmngr.c - LDAP access
  *	Copyright (C) 2002 Klarälvdalens Datakonsult AB
- *      Copyright (C) 2003, 2004 g10 Code GmbH
+ *      Copyright (C) 2003, 2004, 2006 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -47,6 +47,8 @@
 #include "crlcache.h"
 #include "crlfetch.h"
 #include "misc.h"
+#include "i18n.h"
+
 
 enum cmd_and_opt_values 
 { aNull = 0,
@@ -64,6 +66,7 @@ enum cmd_and_opt_values
   aShutdown,
   aFlush,
   aGPGConfList,
+  aGPGConfTest,
 
   oOptions,
   oDebug,
@@ -90,6 +93,8 @@ enum cmd_and_opt_values
   oLDAPAddServers,
   oOCSPResponder,
   oOCSPSigner,
+  oOCSPMaxClockSkew,
+  oOCSPCurrentPeriod,
   oMaxReplies,
   oFakedSystemTime,
   oForce,
@@ -113,6 +118,7 @@ static ARGPARSE_OPTS opts[] = {
   { aShutdown, "shutdown",  256, N_("shutdown the dirmngr")},
   { aFlush,    "flush",     256, N_("flush the cache")},
   { aGPGConfList, "gpgconf-list", 256, "@" },
+  { aGPGConfTest, "gpgconf-test", 256, "@" },
 
   { 301, NULL, 0, N_("@\nOptions:\n ") },
 
@@ -124,7 +130,7 @@ static ARGPARSE_OPTS opts[] = {
   { oDebugLevel, "debug-level",2,
                                N_("|LEVEL|set the debugging level to LEVEL")},
   { oNoDetach, "no-detach" ,0, N_("do not detach from the console")},
-  { oLogFile,  "log-file"  ,2, N_("|FILE|write logs to FILE")},
+  { oLogFile,  "log-file"  ,2, N_("|FILE|write server mode logs to FILE")},
   { oBatch   , "batch"     ,0, N_("run without asking a user")},
   { oForce,    "force"     ,0, N_("force loading of outdated CRLs")},
   { oAllowOCSP, "allow-ocsp",0,N_("allow sending OCSP requests")},
@@ -156,6 +162,8 @@ static ARGPARSE_OPTS opts[] = {
 
   { oOCSPResponder, "ocsp-responder", 2, N_("|URL|use OCSP responder at URL")},
   { oOCSPSigner, "ocsp-signer", 2, N_("|FPR|OCSP response signed by FPR")}, 
+  { oOCSPMaxClockSkew, "ocsp-max-clock-skew", 1, "@" },
+  { oOCSPCurrentPeriod, "ocsp-current-period", 1, "@" },
 
   { oMaxReplies, "max-replies", 1,
     N_("|N|do not return more than N items in one query")},
@@ -190,8 +198,6 @@ static int cleanup_socket;
 /* Keep track of the current log file so that we can avoid updating
    the log file after a SIGHUP if it didn't changed. Malloced. */
 static char *current_logfile;
-/* Name of a config file, which will be reread on a HUP if it is not NULL. */
-static char *config_filename;
 /* Helper to implement --debug-level. */
 static const char *debug_level;
 /* Flag indicating that a shutdown has been requested.  */
@@ -212,28 +218,28 @@ GCRY_THREAD_OPTION_PTH_IMPL;
 static const char *
 my_strusage( int level )
 {
-    const char *p;
-    switch( level ) {
-      case 11: p = "dirmngr";
-	break;
-      case 13: p = VERSION; break;
-      case 14: p = "Copyright (C) 2004 g10 Code GmbH"; break;
-      case 17: p = PRINTABLE_OS_NAME; break;
-      case 19: p =
-	    _("Please report bugs to <gpa-dev@gnupg.org>.\n");
-	break;
-      case 1:
-      case 40:	p =
-	    _("Usage: dirmngr [options] (-h for help)");
-	break;
-      case 41:	p =
-	    _("Syntax: dirmngr [options] [command [args]]\n"
-	      "LDAP and OCSP access for GnuPG\n");
-	break;
-
-      default:	p = NULL;
+  const char *p;
+  switch ( level ) 
+    {
+    case 11: p = "dirmngr";
+      break;
+    case 13: p = VERSION; break;
+    case 14: p = "Copyright (C) " COPYRIGHT_YEAR_NAME; break;
+    case 17: p = PRINTABLE_OS_NAME; break;
+      /* TRANSLATORS: @EMAIL will get replaced by the actual bug
+         reporting address. */
+    case 19: p = _("Please report bugs to <@EMAIL@>.\n"); break;
+    case 49: p = PACKAGE_BUGREPORT; break;
+    case 1:
+    case 40: p = _("Usage: dirmngr [options] (-h for help)");
+      break;
+    case 41: p = _("Syntax: dirmngr [options] [command [args]]\n"
+                   "LDAP and OCSP access for GnuPG\n");
+      break;
+      
+    default: p = NULL;
     }
-    return p;
+  return p;
 }
 
 
@@ -401,6 +407,8 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.allow_ocsp = 0;
       opt.ocsp_responder = NULL;
       opt.ocsp_signer = NULL; 
+      opt.ocsp_max_clock_skew = 10 * 60;      /* 10 minutes.  */
+      opt.ocsp_current_period = 3 * 60 * 60;  /* 3 hours. */
       opt.max_replies = DEFAULT_MAX_REPLIES;
       return 1;
     }
@@ -445,6 +453,8 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
     case oAllowOCSP: opt.allow_ocsp = 1; break;
     case oOCSPResponder: opt.ocsp_responder = pargs->r.ret_str; break;
     case oOCSPSigner:    opt.ocsp_signer = pargs->r.ret_str; break;
+    case oOCSPMaxClockSkew: opt.ocsp_max_clock_skew = pargs->r.ret_int; break;
+    case oOCSPCurrentPeriod: opt.ocsp_current_period = pargs->r.ret_int; break;
 
     case oMaxReplies: opt.max_replies = pargs->r.ret_int; break;
 
@@ -519,6 +529,7 @@ main (int argc, char **argv )
   assuan_set_malloc_hooks (gcry_malloc, gcry_realloc, gcry_free);
   assuan_set_assuan_log_stream (log_get_stream ());
   assuan_set_assuan_log_prefix (log_get_prefix (NULL));
+  assuan_set_assuan_err_source (GPG_ERR_SOURCE_DEFAULT);
 
   /* Setup I18N. */
   i18n_init();
@@ -541,7 +552,7 @@ main (int argc, char **argv )
   /* Reset rereadable options to default values. */
   parse_rereadable_options (NULL, 0); 
 
-  /* LDAP defaults */
+  /* LDAP defaults.  */
   opt.add_new_ldapservers = 0;
   opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
 
@@ -635,6 +646,7 @@ main (int argc, char **argv )
 	case aLoadCRL: 
         case aFetchCRL:
 	case aGPGConfList:
+	case aGPGConfTest:
           cmd = pargs.r_opt;
           break;
 
@@ -686,7 +698,7 @@ main (int argc, char **argv )
       fclose( configfp );
       configfp = NULL;
       /* Keep a copy of the name so that it can be read on SIGHUP. */
-      config_filename = configname;
+      opt.config_filename = configname;
       configname = NULL;
       goto next_pass;
     }
@@ -986,6 +998,8 @@ main (int argc, char **argv )
         wrong_args ("--flush");
       rc = crl_cache_flush();
     }
+  else if (cmd == aGPGConfTest)
+    dirmngr_exit (0);
   else if (cmd == aGPGConfList)
     {
       char *filename;
@@ -1007,11 +1021,12 @@ main (int argc, char **argv )
 
       /* First the configuration file.  This is not an option, but it
 	 is vital information for GPG Conf.  */
-      if (!config_filename)
-        config_filename = make_filename (opt.homedir, "dirmngr.conf", NULL );
+      if (!opt.config_filename)
+        opt.config_filename = make_filename (opt.homedir,
+                                             "dirmngr.conf", NULL );
 
       printf ("gpgconf-dirmngr.conf:%lu:\"%s\n",
-              GC_OPT_FLAG_DEFAULT, config_filename);
+              GC_OPT_FLAG_DEFAULT, opt.config_filename);
 
       printf ("verbose:%lu:\n", GC_OPT_FLAG_NONE);
       printf ("quiet:%lu:\n", GC_OPT_FLAG_NONE);
@@ -1269,14 +1284,14 @@ reread_configuration (void)
   unsigned int configlineno = 0;
   int dummy;
 
-  if (!config_filename)
+  if (!opt.config_filename)
     return; /* No config file. */
 
-  fp = fopen (config_filename, "r");
+  fp = fopen (opt.config_filename, "r");
   if (!fp)
     {
       log_error (_("option file `%s': %s\n"),
-                 config_filename, strerror(errno) );
+                 opt.config_filename, strerror(errno) );
       return;
     }
 
@@ -1286,7 +1301,7 @@ reread_configuration (void)
   dummy = 0;
   pargs.argc = &dummy;
   pargs.flags = 1;  /* do not remove the args */
-  while (optfile_parse (fp, config_filename, &configlineno, &pargs, opts) )
+  while (optfile_parse (fp, opt.config_filename, &configlineno, &pargs, opts) )
     {
       if (pargs.r_opt < -1)
         pargs.err = 1; /* Print a warning. */
