@@ -36,6 +36,20 @@
 #define MAX_RESPONSE_SIZE 65536
 
 
+static const char oidstr_ocsp[] = "1.3.6.1.5.5.7.48.1";
+
+
+/* Telesec attribute used to implement a positive confirmation. 
+
+   CertHash ::= SEQUENCE {
+      HashAlgorithm    AlgorithmIdentifier,
+      certificateHash OCTET STRING }
+ */
+static const char oidstr_certHash[] = "1.3.36.8.3.13";
+
+
+
+
 /* Read from FP and return a newly allocated buffer in R_BUFFER with the
    entire data read from FP. */
 static gpg_error_t
@@ -111,9 +125,11 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
   gpg_error_t err;
   unsigned char *request, *response;
   size_t requestlen, responselen;
-  struct http_context_s http;
+  http_t http;
   ksba_ocsp_response_status_t response_status;
   const char *t;
+  int redirects_left = 2;
+  char *free_this = NULL;
 
   if (opt.disable_http)
     {
@@ -146,53 +162,91 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
       return err;
     }
 
-  err = http_open (&http, HTTP_REQ_POST, url,
-                   opt.honor_http_proxy? HTTP_FLAG_TRY_PROXY:0,
-                   opt.http_proxy);
+ once_more:
+  err = http_open (&http, HTTP_REQ_POST, url, NULL,
+                   (opt.honor_http_proxy? HTTP_FLAG_TRY_PROXY:0)
+                   |HTTP_FLAG_NEED_HEADER,
+                   opt.http_proxy,
+                   NULL);
   if (err)
     {
       log_error (_("error connecting to `%s': %s\n"), url, gpg_strerror (err));
+      xfree (free_this);
       return err;
     }
 
-  fprintf (http.fp_write,
+  fprintf (http_get_write_ptr (http),
            "Content-Type: application/ocsp-request\r\n"
            "Content-Length: %lu\r\n", 
            (unsigned long)requestlen );
-  http_start_data (&http);
-  if (fwrite (request, requestlen, 1, http.fp_write) != 1)
+  http_start_data (http);
+  if (fwrite (request, requestlen, 1, http_get_write_ptr (http)) != 1)
     {
       err = gpg_error_from_errno (errno);
       log_error ("error sending request to `%s': %s\n", url, strerror (errno));
-      http_close (&http, 0);
+      http_close (http, 0);
       xfree (request);
+      xfree (free_this);
       return err;
     }
   xfree (request);
   request = NULL;
 
-  err = http_wait_response (&http, NULL);
-  if (err || http.status_code != 200)
+  err = http_wait_response (http);
+  if (err || http_get_status_code (http) != 200)
     {
       if (err)
         log_error (_("error reading HTTP response for `%s': %s\n"),
                    url, gpg_strerror (err));
       else
         {
-          log_error (_("error accessing `%s': http status %u\n"),
-                     url, http.status_code);
-          err = gpg_error (GPG_ERR_NO_DATA);
+          switch (http_get_status_code (http))
+            {
+            case 301:
+            case 302:
+              {
+                const char *s = http_get_header (http, "Location");
+                
+                log_info (_("URL `%s' redirected to `%s' (%u)\n"),
+                          url, s?s:"[none]", http_get_status_code (http));
+                if (s && *s && redirects_left-- )
+                  {
+                    xfree (free_this); url = NULL;
+                    free_this = xtrystrdup (s);
+                    if (!free_this)
+                      err = gpg_error_from_errno (errno);
+                    else
+                      {
+                        url = free_this;
+                        http_close (http, 0);
+                        goto once_more;
+                      }
+                  }
+                else
+                  err = gpg_error (GPG_ERR_NO_DATA);
+                log_error (_("too many redirections\n"));
+              }
+              break;
+
+            default:
+              log_error (_("error accessing `%s': http status %u\n"),
+                         url, http_get_status_code (http));
+              err = gpg_error (GPG_ERR_NO_DATA);
+              break;
+            }
         }
-      http_close (&http, 0);
+      http_close (http, 0);
+      xfree (free_this);
       return err;
     }
 
-  err = read_response (http.fp_read, &response, &responselen);
-  http_close (&http, 0);
+  err = read_response (http_get_read_ptr (http), &response, &responselen);
+  http_close (http, 0);
   if (err)
     {
       log_error (_("error reading HTTP response for `%s': %s\n"),
                  url, gpg_strerror (err));
+      xfree (free_this);
       return err;
     }
 
@@ -203,6 +257,7 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
       log_error (_("error parsing OCSP response for `%s': %s\n"),
                  url, gpg_strerror (err));
       xfree (response);
+      xfree (free_this);
       return err;
     }
 
@@ -237,18 +292,34 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
     }
 
   xfree (response);
+  xfree (free_this);
   return err;
 }
 
 
-/* Validate that CERT is indeed valid to sign an OCSP response. */
+/* Validate that CERT is indeed valid to sign an OCSP response. If
+   signer_fpr is not NULL we simply check that CERT matches this
+   fingerprint. */
 static gpg_error_t
-validate_responder_cert (ctrl_t ctrl, ksba_cert_t cert)
+validate_responder_cert (ctrl_t ctrl, ksba_cert_t cert, const char *signer_fpr)
 {
   gpg_error_t err;
   char *fpr;
 
-  if (opt.system_daemon)
+  if (signer_fpr)
+    {
+      fpr = (strchr (signer_fpr, ':')
+             ? get_fingerprint_hexstring_colon (cert)
+             : get_fingerprint_hexstring (cert));
+      if (ascii_strcasecmp (signer_fpr, fpr))
+        {
+          log_error (_("not signed by default OCSP signer certificate"));
+          err = gpg_error (GPG_ERR_BAD_CA_CERT);
+        }
+      else
+        err = 0;
+    }
+  else if (opt.system_daemon)
     {
       err = validate_cert_chain (ctrl, cert, NULL, VALIDATE_MODE_OCSP);
     }
@@ -283,7 +354,7 @@ validate_responder_cert (ctrl_t ctrl, ksba_cert_t cert)
 /* Helper for check_signature. */
 static int
 check_signature_core (ctrl_t ctrl, ksba_cert_t cert, gcry_sexp_t s_sig,
-                      gcry_sexp_t s_hash)
+                      gcry_sexp_t s_hash, const char *signer_fpr)
 {
   gpg_error_t err;
   ksba_sexp_t pubkey;
@@ -298,7 +369,7 @@ check_signature_core (ctrl_t ctrl, ksba_cert_t cert, gcry_sexp_t s_sig,
   if (!err)
     err = gcry_pk_verify (s_sig, s_hash, s_pkey);
   if (!err)
-    err = validate_responder_cert (ctrl, cert);
+    err = validate_responder_cert (ctrl, cert, signer_fpr);
   if (!err)
     {
       gcry_sexp_release (s_pkey);
@@ -344,16 +415,19 @@ check_signature (ctrl_t ctrl,
       return err;
     }
 
+  /* Get rid of old OCSP specific certificate references. */
+  release_ctrl_ocsp_certs (ctrl);
+
   if (signer_fpr)
     {
-      /* We should use the default OCSP reponder's certificate.  Get
-         it from the fingerprint. */
+      /* We use the default OCSP responder's certificate.  Get it from
+         the fingerprint. */
       cert = get_cert_byhexfpr (signer_fpr);
       if (!cert)
         cert = get_cert_local (ctrl, signer_fpr);
       if (cert)
         {
-          err = check_signature_core (ctrl, cert, s_sig, s_hash);
+          err = check_signature_core (ctrl, cert, s_sig, s_hash, signer_fpr);
           ksba_cert_release (cert);
           cert = NULL;
           if (!err)
@@ -365,16 +439,59 @@ check_signature (ctrl_t ctrl,
     }
   else
     {
-      /* As of now we rely on having a valid certificate in the response.
-         Obviously this may not be true in all cases and thus we should
-         get the responder ID and try to locate the certificate by other
-         means too. */
+      char *name;
+      ksba_sexp_t keyid;
+
+      /* Put all certificates included in the response into the cache
+         and setup a list of those certificate which will later be
+         preferred used when locating certificates.  */
       for (cert_idx=0; (cert = ksba_ocsp_get_cert (ocsp, cert_idx));
            cert_idx++)
         {
-          err = check_signature_core (ctrl, cert, s_sig, s_hash);
+          cert_ref_t cref;
+
+          cref = xtrymalloc (sizeof *cref);
+          if (!cref)
+            log_error (_("allocating list item failed: %s\n"),
+                       gcry_strerror (err));
+          else if (!cache_cert_silent (cert, &cref->fpr))
+            {
+              cref->next = ctrl->ocsp_certs;
+              ctrl->ocsp_certs = cref;
+            }
+          else
+            xfree (cref);
+        }
+
+      /* Get the certificate by means of the respinder ID. */
+      err = ksba_ocsp_get_responder_id (ocsp, &name, &keyid);
+      if (err)
+        {
+          log_error (_("error getting responder ID: %s\n"),
+                     gcry_strerror (err));
+          return err;
+        }
+      cert = find_cert_bysubject (ctrl, name, keyid);
+      if (!cert)
+        {
+          log_error ("responder certificate ");
+          if (name)
+            log_printf ("`/%s' ", name);
+          if (keyid)
+            {
+              log_printf ("{");
+              dump_serial (keyid);
+              log_printf ("} ");
+            }
+          log_printf ("not found\n");
+        }
+      ksba_free (name);
+      ksba_free (keyid);
+
+      if (cert)
+        {
+          err = check_signature_core (ctrl, cert, s_sig, s_hash, NULL);
           ksba_cert_release (cert);
-          cert = NULL;
           if (!err)
             {
               gcry_sexp_release (s_hash);
@@ -390,7 +507,7 @@ check_signature (ctrl_t ctrl,
 
 
 /* Check whether the certificate either given by fingerprint CERT_FPR
-   or directly through a the CERT object is valid by running an OCSP
+   or directly through the CERT object is valid by running an OCSP
    transaction.  */
 gpg_error_t
 ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr)
@@ -465,7 +582,7 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr)
          && !(err=ksba_cert_get_authority_info_access (cert, idx,
                                                        &oid, &name)); idx++)
     {
-      if ( !strcmp (oid, "1.3.6.1.5.5.7.48.1") )
+      if ( !strcmp (oid, oidstr_ocsp) )
         {
           for (i=0; !url && ksba_name_enum (name, i); i++)
             {
@@ -548,6 +665,24 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr)
                  gpg_strerror (err));
       goto leave;
     }
+
+  /* In case the certificate has been revoked, we better invalidate
+     our cached validation status. */
+  if (status == KSBA_STATUS_REVOKED)
+    {
+      time_t validated_at = 0; /* That is: No cached validation available. */
+      err = ksba_cert_set_user_data (cert, "validated_at",
+                                     &validated_at, sizeof (validated_at));
+      if (err)
+        {
+          log_error ("set_user_data(validated_at) failed: %s\n",
+                     gpg_strerror (err)); 
+          err = 0; /* The certificate is anyway revoked, and that is a
+                      more important message than the failure of our
+                      cache. */
+        }
+    }
+
   
   if (opt.verbose)
     {
@@ -586,9 +721,9 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr)
     err = gpg_error (GPG_ERR_GENERAL);
 
   get_isotime (current_time);
-  /* Allow for 10 minutes of clock skew. Note, that NEXT_UPDATE is
+  /* Allow for some clock skew. Note, that NEXT_UPDATE is
      optional. */
-  add_isotime (current_time, 10 * 60);
+  add_isotime (current_time, opt.ocsp_max_clock_skew);
   if (*next_update && strcmp (next_update, current_time) < 0 )
     {
       log_error (_("OCSP responder returned an too old status\n"));
@@ -596,11 +731,9 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr)
       if (!err)
         err = gpg_error (GPG_ERR_TIME_CONFLICT);
     }
-  /* Check that THIS_UPDATE is not too far back in the past.  We
-     currently use 3 hours (the extra 10 minutes are for the time
-     adjust above). */
+  /* Check that THIS_UPDATE is not too far back in the past. */
   copy_time (tmp_time, this_update);
-  add_isotime (this_update, 3 * 60 * 60 + 10 * 60);
+  add_isotime (this_update, opt.ocsp_current_period + opt.ocsp_max_clock_skew);
   if (!*this_update || strcmp (this_update, current_time) < 0 )
     {
       log_error (_("OCSP responder returned a non-current status\n"));
@@ -621,3 +754,14 @@ ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr)
 }
 
 
+/* Release the list of OCSP certificates hold in the CTRL object. */
+void
+release_ctrl_ocsp_certs (ctrl_t ctrl)
+{
+  while (ctrl->ocsp_certs)
+    {
+      cert_ref_t tmp = ctrl->ocsp_certs->next;
+      xfree (ctrl->ocsp_certs);
+      ctrl->ocsp_certs = tmp;
+    }
+}
