@@ -45,38 +45,17 @@
    something reasonable. */
 #define MAX_CERT_LENGTH (8*1024)
 
-#define PARM_ERROR(t) assuan_set_error (ctx, ASSUAN_Parameter_Error, (t))
+#define PARM_ERROR(t) assuan_set_error (ctx, \
+                                        gpg_error (GPG_ERR_ASS_PARAMETER), (t))
 
 
 /* Data used to associate an Assuan context with local server data */
-struct server_local_s {
-  ASSUAN_CONTEXT assuan_ctx;
+struct server_local_s 
+{
+  assuan_context_t assuan_ctx;
 };
 
 
-/* Map GPG_ERR_xx error codes to Assuan status codes */
-static int
-map_to_assuan_status (int rc)
-{
-  gpg_err_code_t   ec = gpg_err_code (rc);
-  gpg_err_source_t es = gpg_err_source (rc);
-
-  if (!rc)
-    return 0;
-  if (!es)
-    {
-      es = GPG_ERR_SOURCE_USER_4; /*  This should not happen, but we
-                                      need to make sure to pass a new
-                                      Assuan error code along. */
-      log_debug ("map_to_assuan_status called with no error source\n");
-    }
-
-  if (ec == -1)
-    ec = GPG_ERR_NO_DATA;  /* That used to be ASSUAN_No_Data_Available. */
-
-  return gpg_err_make (es, ec);
-}
-
 
 
 /* Copy the % and + escaped string S into the buffer D and replace the
@@ -130,7 +109,7 @@ do_get_cert_local (ctrl_t ctrl, const char *name, const char *command)
   if (rc)
     {
       log_error (_("assuan_inquire(%s) failed: %s\n"),
-                 command, assuan_strerror (rc));
+                 command, gpg_strerror (rc));
       return NULL;
     }
   
@@ -242,7 +221,7 @@ get_cert_local_ski (ctrl_t ctrl, const char *name, ksba_sexp_t keyid)
   if (rc)
     {
       log_error (_("assuan_inquire(%s) failed: %s\n"), "SENDCERT_SKI",
-                 assuan_strerror (rc));
+                 gpg_strerror (rc));
       return NULL;
     }
   
@@ -277,14 +256,13 @@ inquire_cert_and_load_crl (assuan_context_t ctx)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
-  assuan_error_t ae;
   unsigned char *value = NULL;
   size_t valuelen; 
   ksba_cert_t cert = NULL;
 
-  ae = assuan_inquire( ctx, "SENDCERT", &value, &valuelen, 0);
-  if (ae)
-    return map_assuan_err (ae);
+  err = assuan_inquire( ctx, "SENDCERT", &value, &valuelen, 0);
+  if (err)
+    return err;
 
 /*   { */
 /*     FILE *fp = fopen ("foo.der", "r"); */
@@ -315,7 +293,7 @@ inquire_cert_and_load_crl (assuan_context_t ctx)
 
 /* Handle OPTION commands. */
 static int
-option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
+option_handler (assuan_context_t ctx, const char *key, const char *value)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
 
@@ -325,7 +303,7 @@ option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
       ctrl->force_crl_refresh = i;
     }
   else
-    return ASSUAN_Invalid_Option;
+    return gpg_error (GPG_ERR_UNKNOWN_OPTION);
 
   return 0;
 }
@@ -427,7 +405,7 @@ cmd_isvalid (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "ISVALID", gpg_strerror (err));
   xfree (issuerhash);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -503,16 +481,14 @@ cmd_checkcrl (assuan_context_t ctx, char *line)
     {
       /* We do not have this certificate yet or the fingerprint has
          not been given.  Inquire it from the client.  */
-      assuan_error_t ae;
       unsigned char *value = NULL;
       size_t valuelen; 
       
-      ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+      err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                            &value, &valuelen, MAX_CERT_LENGTH);
-      if (ae)
+      if (err)
         {
-          log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-          err = map_assuan_err (ae);
+          log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
           goto leave;
         }
   
@@ -543,7 +519,7 @@ cmd_checkcrl (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "CHECKCRL", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -585,16 +561,14 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
     {
       /* We do not have this certificate yet or the fingerprint has
          not been given.  Inquire it from the client.  */
-      assuan_error_t ae;
       unsigned char *value = NULL;
       size_t valuelen; 
       
-      ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+      err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                            &value, &valuelen, MAX_CERT_LENGTH);
-      if (ae)
+      if (err)
         {
-          log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-          err = map_assuan_err (ae);
+          log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
           goto leave;
         }
   
@@ -622,7 +596,7 @@ cmd_checkocsp (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "CHECKOCSP", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -640,7 +614,6 @@ cmd_lookup (assuan_context_t ctx, char *line)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err = 0;
-  assuan_error_t ae;
   char *p;
   strlist_t sl, list = NULL;
   int truncated = 0, truncation_forced = 0;
@@ -649,6 +622,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
   size_t valuelen; 
   ldap_server_t ldapserver;
   cert_fetch_context_t fetch_context;
+  int any_no_data = 0;
 
   /* Break the line down into an STRLIST */
   for (p=line; *p; line = p)
@@ -686,6 +660,14 @@ cmd_lookup (assuan_context_t ctx, char *line)
 
       /* Fetch certificates matching pattern */
       err = start_cert_fetch (ctrl, &fetch_context, list, ldapserver);
+      if ( gpg_err_code (err) == GPG_ERR_NO_DATA )
+        {
+          if (DBG_LOOKUP)
+            log_debug ("cmd_lookup: no data\n");
+          err = 0;
+          any_no_data = 1;
+          continue;
+        }
       if (err)
         {
           log_error (_("start_cert_fetch failed: %s\n"), gpg_strerror (err));
@@ -697,6 +679,12 @@ cmd_lookup (assuan_context_t ctx, char *line)
         {
           xfree (value); value = NULL;
           err = fetch_next_cert (fetch_context, &value, &valuelen);
+          if (gpg_err_code (err) == GPG_ERR_NO_DATA )
+            {
+              err = 0;
+              any_no_data = 1;
+              break; /* Ready. */
+            }
           if (gpg_err_code (err) == GPG_ERR_TRUNCATED)
             {
               truncated = 1;
@@ -727,15 +715,14 @@ cmd_lookup (assuan_context_t ctx, char *line)
           
           /* Send the data, flush the buffer and then send an END line
              as a certificate delimiter. */
-          ae = assuan_send_data (ctx, value, valuelen);      
-          if (!ae)
-            ae = assuan_send_data (ctx, NULL, 0);
-          if (!ae)
-            ae = assuan_write_line (ctx, "END");
-          if (ae) 
+          err = assuan_send_data (ctx, value, valuelen);      
+          if (!err)
+            err = assuan_send_data (ctx, NULL, 0);
+          if (!err)
+            err = assuan_write_line (ctx, "END");
+          if (err) 
             {
-              log_error (_("error sending data: %s\n"), assuan_strerror (ae));
-              err = map_assuan_err (ae);
+              log_error (_("error sending data: %s\n"), gpg_strerror (err));
               end_cert_fetch (fetch_context);
               goto leave;
             }
@@ -758,11 +745,14 @@ cmd_lookup (assuan_context_t ctx, char *line)
       assuan_write_status (ctx, "TRUNCATED", str);    
     }
 
+  if (!err && !count && any_no_data)
+    err = gpg_error (GPG_ERR_NO_DATA);
+
  leave:
   if (err)
     log_error (_("command %s failed: %s\n"), "LOOKUP", gpg_strerror (err));
   free_strlist (list);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -789,7 +779,7 @@ cmd_loadcrl (assuan_context_t ctx, char *line)
   xfree (buf);
   if (err)
     log_error (_("command %s failed: %s\n"), "LOADCRL", gpg_strerror (err));
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
@@ -813,14 +803,14 @@ cmd_listcrls (assuan_context_t ctx, char *line)
   err = crl_cache_list (fp);
   if (err)
     log_error (_("command %s failed: %s\n"), "LISTCRLS", gpg_strerror (err));
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
 /* CACHECERT 
 
    Put a certificate into the internal cache.  This command might be
-   usedful if a client knows in advance certificates required for a
+   useful if a client knows in advance certificates required for a
    test and wnats to make sure they get added to the internal cache.
    It is also helpful for debugging.  To get the actual certificate,
    this command immediately inquires it using
@@ -836,16 +826,14 @@ cmd_cachecert (assuan_context_t ctx, char *line)
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
   ksba_cert_t cert = NULL;
-  assuan_error_t ae;
   unsigned char *value = NULL;
   size_t valuelen; 
       
-  ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+  err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                        &value, &valuelen, MAX_CERT_LENGTH);
-  if (ae)
+  if (err)
     {
-      log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-      err = map_assuan_err (ae);
+      log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
       goto leave;
     }
   
@@ -867,13 +855,13 @@ cmd_cachecert (assuan_context_t ctx, char *line)
   if (err)
     log_error (_("command %s failed: %s\n"), "CACHECERT", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
 /* VALIDATE 
 
-   Validate a certificate using the certificate validationj fucntion
+   Validate a certificate using the certificate validation fucntion
    used internally by dirmngr.  This command is only useful for
    debugging.  To get the actual certificate, this command immediately
    inquires it using
@@ -889,16 +877,14 @@ cmd_validate (assuan_context_t ctx, char *line)
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
   ksba_cert_t cert = NULL;
-  assuan_error_t ae;
   unsigned char *value = NULL;
   size_t valuelen; 
-      
-  ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
+    
+  err = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                        &value, &valuelen, MAX_CERT_LENGTH);
-  if (ae)
+  if (err)
     {
-      log_error (_("assuan_inquire failed: %s\n"), assuan_strerror (ae));
-      err = map_assuan_err (ae);
+      log_error (_("assuan_inquire failed: %s\n"), gpg_strerror (err));
       goto leave;
     }
   
@@ -914,24 +900,40 @@ cmd_validate (assuan_context_t ctx, char *line)
   if(err)
     goto leave;
 
-  err = validate_cert_chain (ctrl, cert, NULL, VALIDATE_MODE_CRL_RECURSIVE);
+  /* If we have this certificate already in our cache, use the cached
+     version for validation because this will take care of any cached
+     results. */
+  { 
+    unsigned char fpr[20];
+    ksba_cert_t tmpcert;
+
+    cert_compute_fpr (cert, fpr);
+    tmpcert = get_cert_byfpr (fpr);
+    if (tmpcert)
+      {
+        ksba_cert_release (cert);
+        cert = tmpcert;
+      }
+  }
+
+  err = validate_cert_chain (ctrl, cert, NULL, VALIDATE_MODE_CERT);
 
  leave:
   if (err)
     log_error (_("command %s failed: %s\n"), "VALIDATE", gpg_strerror (err));
   ksba_cert_release (cert);
-  return map_to_assuan_status (err);
+  return err;
 }
 
 
 
 /* Tell the assuan library about our commands. */
 static int
-register_commands (ASSUAN_CONTEXT ctx)
+register_commands (assuan_context_t ctx)
 {
   static struct {
     const char *name;
-    int (*handler)(ASSUAN_CONTEXT, char *line);
+    int (*handler)(assuan_context_t, char *line);
   } table[] = {
     { "ISVALID",    cmd_isvalid },
     { "CHECKCRL",   cmd_checkcrl },
@@ -962,6 +964,8 @@ register_commands (ASSUAN_CONTEXT ctx)
 void
 start_command_handler (int fd)
 {
+  static const char hello[] = "Dirmngr " VERSION " at your service";
+  static char *hello_line;
   int rc;
   assuan_context_t ctx;
   ctrl_t ctrl;
@@ -989,13 +993,13 @@ start_command_handler (int fd)
     }
   else
     {
-      rc = assuan_init_connected_socket_server (&ctx, fd);
+      rc = assuan_init_socket_server_ext (&ctx, fd, 2);
     }
 
   if (rc)
     {
       log_error (_("failed to initialize the server: %s\n"),
-                 assuan_strerror(rc));
+                 gpg_strerror(rc));
       dirmngr_exit (2);
     }
 
@@ -1003,10 +1007,32 @@ start_command_handler (int fd)
   if (rc)
     {
       log_error (_("failed to the register commands with Assuan: %s\n"),
-                 assuan_strerror(rc));
+                 gpg_strerror(rc));
       dirmngr_exit (2);
     }
-  assuan_set_hello_line (ctx, "Dirmngr " VERSION " at your service");
+
+
+  if (!hello_line)
+    {
+      size_t n;
+      const char *cfgname;
+
+      cfgname = opt.config_filename? opt.config_filename : "[none]";
+
+      n = (30 + strlen (opt.homedir) + strlen (cfgname)
+           + strlen (hello) + 1);
+      hello_line = xmalloc (n+1);
+      snprintf (hello_line, n,
+                "Home: %s\n"
+                "Config: %s\n"
+                "%s",
+                opt.homedir,
+                cfgname,
+                hello);
+      hello_line[n] = 0;
+    }
+
+  assuan_set_hello_line (ctx, hello_line);
   assuan_register_option_handler (ctx, option_handler);
 
   ctrl->server_local->assuan_ctx = ctx;
@@ -1022,14 +1048,25 @@ start_command_handler (int fd)
         break;
       if (rc)
         {
-          log_info (_("Assuan accept problem: %s\n"), assuan_strerror (rc));
+          log_info (_("Assuan accept problem: %s\n"), gpg_strerror (rc));
           break;
         }
 
+      if (opt.verbose)
+        {
+          pid_t apid;
+          uid_t auid;
+          gid_t agid;
+
+          if (!assuan_get_peercred (ctx, &apid, &auid, &agid))
+            log_info ("connection from process %ld (%ld:%ld)\n",
+                      (long)apid, (long)auid, (long)agid);
+        }
+
       rc = assuan_process (ctx);
       if (rc)
         {
-          log_info (_("Assuan processing failed: %s\n"), assuan_strerror (rc));
+          log_info (_("Assuan processing failed: %s\n"), gpg_strerror (rc));
           continue;
         }
     }
@@ -1044,6 +1081,7 @@ start_command_handler (int fd)
                ctrl->refcount);
   else
     {
+      release_ctrl_ocsp_certs (ctrl);
       xfree (ctrl->server_local);
       xfree (ctrl);
     }
@@ -1081,7 +1119,7 @@ dirmngr_status (ctrl_t ctrl, const char *keyword, ...)
             *p++ = *text++;
         }
       *p = 0;
-      err = map_assuan_err (assuan_write_status (ctx, keyword, buf));
+      err = assuan_write_status (ctx, keyword, buf);
     }
 
   va_end (arg_ptr);
